@@ -14,11 +14,12 @@ ident = lambda x: x
 
 
 def common(format, data=NotImplemented, obj=NotImplemented, size=NotImplemented, eq=True):
+    if data is not NotImplemented:
+        data2 = data.strip(b"?")
     if data is not NotImplemented and obj is not NotImplemented:
         assert format.parse(data) == obj
         assert format.build(obj) == data
     else:
-        # following are guaranteed by the above
         if obj is not NotImplemented:
             if eq:
                 assert format.parse(format.build(obj)) == obj
@@ -99,6 +100,8 @@ class TestCore(unittest.TestCase):
         common(Int24ul, b"\x01\x02\x03", 0x030201, 3)
         common(Struct(int24=Int24ub), b"\x01\x02\x03", Container(int24=0x010203), 3)
         common(Struct(int24=Int24ul), b"\x01\x02\x03", Container(int24=0x030201), 3)
+        common(Int24sb, b"\xff\xff\xff", -1, 3)
+        common(Int24sl, b"\xff\xff\xff", -1, 3)
 
     def test_varint(self):
         common(VarInt, b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x10", 2**123, SizeofError)
@@ -208,36 +211,10 @@ class TestCore(unittest.TestCase):
         assert PrefixedArray(Byte, Byte).parse(b"\x03\x01\x02\x03") == [1,2,3]
         assert PrefixedArray(Byte, Byte).parse(b"\x00") == []
         assert PrefixedArray(Byte, Byte).build([1,2,3]) == b"\x03\x01\x02\x03"
-        assert raises(PrefixedArray(Byte, Byte).parse, b"") == RangeError
+        assert raises(PrefixedArray(Byte, Byte).parse, b"") == FieldError
         assert raises(PrefixedArray(Byte, Byte).parse, b"\x03\x01") == RangeError
         assert raises(PrefixedArray(Byte, Byte).sizeof) == SizeofError
 
-    def test_prefixedarray_alternative2(self):
-        def PrefixedArray(lengthfield, subcon):
-            return FocusedSeq(1, 
-                "count"/Rebuild(lengthfield, len_(this.items)), 
-                "items"/subcon[this.count],
-            )
-        common(PrefixedArray(Byte,Byte), b"\x02\x0a\x0b", [10,11], SizeofError)
-
-    @pytest.mark.xfail(raises=TypeError, reason="object of type 'instancemethod' has no len()")
-    def test_prefixedarray_alternative3(self):
-        def PrefixedArray(lengthfield, subcon):
-            return FocusedSeq(1, 
-                "count"/Rebuild(lengthfield, lambda ctx: len(ctx.items)), 
-                "items"/Array(lambda ctx: ctx.count, subcon),
-            )
-        common(PrefixedArray(Byte,Byte), b"\x02\x0a\x0b", [10,11], AttributeError)
-
-    def test_prefixedarray_alternative4(self):
-        def PrefixedArray(lengthfield, subcon):
-            return ExprAdapter(
-                "count"/lengthfield >> "items"/subcon[lambda ctx: ctx.count],
-                encoder = lambda obj,ctx: [len(obj), obj],
-                decoder = lambda obj,ctx: obj[1],
-            )
-        common(PrefixedArray(Byte,Byte), b"\x02\x0a\x0b", [10,11], SizeofError)
-
     def test_range(self):
         assert Byte[2:4].parse(b"1234567890") == [49, 50, 51, 52]
         assert Byte[2:4].build([49, 50, 51, 52]) == b"1234"
@@ -278,6 +255,8 @@ class TestCore(unittest.TestCase):
         assert raises(RepeatUntil(obj_ == 9, Byte).parse, b"\x02\x03\x08") == RangeError
         assert raises(RepeatUntil(obj_ == 9, Byte).build, [2,3,8]) == RangeError
         assert raises(RepeatUntil(obj_ == 9, Byte).sizeof) == SizeofError
+        assert RepeatUntil(lambda x,lst,ctx: lst[-2:]==[0,0], Byte).parse(b"\x01\x00\x00\xff") == [1,0,0]
+        assert RepeatUntil(lambda x,lst,ctx: lst[-2:]==[0,0], Byte).build([1,0,0,4]) == b"\x01\x00\x00"
 
     def test_struct(self):
         assert Struct("a" / Int16ul, "b" / Byte).parse(b"\x01\x00\x02") == Container(a=1)(b=2)
@@ -300,10 +279,33 @@ class TestCore(unittest.TestCase):
         assert Struct("a" / Byte, "b" / Int16ub, Embedded("inner" / Struct("c" / Byte, "d" / Byte))).parse(b"\x01\x00\x02\x03\x04") == Container(a=1)(b=2)(c=3)(d=4)
         assert Struct("a" / Byte, "b" / Int16ub, Embedded("inner" / Struct("c" / Byte, "d" / Byte))).build(Container(a=1)(b=2)(c=3)(d=4)) == b"\x01\x00\x02\x03\x04"
 
-    @pytest.mark.xfail(not supportskwordered, reason="ordered kw was introduced in 3.6 and pypy")
+    @pytest.mark.xfail(not supportskwordered, reason="ordered kw was introduced in 3.6")
     def test_struct_kwctor(self):
         common(Struct(a=Byte, b=Byte, c=Byte, d=Byte), b"\x01\x02\x03\x04", Container(a=1,b=2,c=3,d=4), 4)
 
+    def test_struct_sizeof_context_nesting(self):
+        st = Struct(
+            "length"/Byte,
+            "inner"/Struct(
+                "inner_length"/Byte,
+                "data"/Bytes(this._.length + this.inner_length),
+        ))
+        assert st.parse(b"\x03\x02helloXXX") == Container(length=3)(inner=Container(inner_length=2)(data=b"hello"))
+        assert st.sizeof(Container(length=3)(inner=Container(inner_length=2))) == 7
+
+        st = Struct(
+            "a" / Byte,
+            "inner" / Struct(
+                "b" / Byte,
+                If(this._.a == 1, Byte),
+                If(this.b == 1, Byte),
+            ),
+            If(this.a == 1, Byte),
+            If(this.inner.b == 1, Byte),
+        )
+        assert st.sizeof(Container(a=0,inner=Container(b=0))) == 2
+        assert st.sizeof(Container(a=1,inner=Container(b=1))) == 6
+
     def test_sequence(self):
         common(Sequence(Int8ub, Int16ub), b"\x01\x00\x02", [1,2], 3)
         common(Int8ub >> Int16ub, b"\x01\x00\x02", [1,2], 3)
@@ -334,6 +336,18 @@ class TestCore(unittest.TestCase):
         assert RawCopy(Byte).build(dict(value=255)) == b"\xff"
         assert RawCopy(Byte).sizeof() == 1
 
+    def test_rawcopy_issue_289(self):
+        # When you build from a full dict that has all the keys, the if data kicks in, and replaces the context entry with a subset of a dict it had to begin with.
+        st = Struct(
+            "raw" / RawCopy(Struct("x"/Byte, "len"/Byte)),
+            "array" / Byte[this.raw.value.len],
+        )
+        print(st.parse(b"\x01\x02\xff\x00"))
+        print(st.build(dict(raw=dict(value=dict(x=1, len=2)), array=[0xff, 0x01])))
+        print(st.build(st.parse(b"\x01\x02\xff\x00")))
+        # this is not buildable, array is not passed and cannot be deduced from raw data anyway
+        # print(st.build(dict(raw=dict(data=b"\x01\x02\xff\x00"))))
+
     def test_tell(self):
         assert Tell.parse(b"") == 0
         assert Tell.build(None) == b""
@@ -343,13 +357,17 @@ class TestCore(unittest.TestCase):
         assert Struct("a"/Tell, "b"/Byte, "c"/Tell).build(dict(b=255)) == b"\xff"
 
     def test_seek(self):
-        assert (Seek(5) >> Byte).parse(b"01234x") == [5,120]
-        assert (Bytes(10) >> Seek(5) >> Byte).build([b"0123456789",None,255]) == b"01234\xff6789"
         assert Seek(5).parse(b"") == 5
         assert Seek(5).build(None) == b""
-        assert raises(Seek(5).sizeof) == SizeofError
-
+        assert (Seek(5) >> Byte).parse(b"01234x") == [5,120]
+        assert (Seek(5) >> Byte).build([5,255]) == b"\x00\x00\x00\x00\x00\xff"
+        assert (Bytes(10) >> Seek(5) >> Byte).parse(b"0123456789") == [b"0123456789",5,ord('5')]
+        assert (Bytes(10) >> Seek(5) >> Byte).build([b"0123456789",None,255]) == b"01234\xff6789"
+        assert Struct("data"/Bytes(10), Seek(5), "addin"/Byte).parse(b"0123456789") == Container(data=b"0123456789")(addin=53)
+        assert Struct("data"/Bytes(10), Seek(5), "addin"/Byte).build(dict(data=b"0123456789",addin=53)) == b"01234\x356789"
         assert (Seek(10,1) >> Seek(-5,1) >> Bytes(1)).parse(b"0123456789") == [10,5,b"5"]
+        assert (Seek(10,1) >> Seek(-5,1) >> Bytes(1)).build([None,None,255]) == b"\x00\x00\x00\x00\x00\xff"
+        assert raises(Seek(5).sizeof) == SizeofError
 
     def test_pass(self):
         common(Pass, b"", None, 0)
@@ -394,12 +412,12 @@ class TestCore(unittest.TestCase):
         assert raises(Switch(lambda ctx: 5, {}).sizeof) == SwitchError
 
     def test_ifthenelse(self):
-        common(IfThenElse(True_,  Int8ub, Int16ub), b"\x01", 1, 1)
-        common(IfThenElse(False_, Int8ub, Int16ub), b"\x00\x01", 1, 2)
+        common(IfThenElse(True,  Int8ub, Int16ub), b"\x01", 1, 1)
+        common(IfThenElse(False, Int8ub, Int16ub), b"\x00\x01", 1, 2)
 
     def test_if(self):
-        common(If(True_,  Int8ub), b"\x01", 1, 1)
-        common(If(False_, Int8ub), b"", None, 0)
+        common(If(True,  Int8ub), b"\x01", 1, 1)
+        common(If(False, Int8ub), b"", None, 0)
 
     def test_padding(self):
         assert Padding(4).parse(b"\x00\x00\x00\x00") == None
@@ -478,14 +496,6 @@ class TestCore(unittest.TestCase):
     def test_renamed(self):
         common(Struct(Renamed("new", Renamed("old", Byte))), b"\x01", Container(new=1), 1)
 
-    def test_alias(self):
-        assert Alias("b","a").parse(b"",Container(a=1)) == 1
-        assert Alias("b","a").build(None,Container(a=1)) == b""
-        assert Alias("b","a").sizeof() == 0
-        assert Struct("a"/Byte, Alias("b","a")).parse(b"\x01") == Container(a=1)(b=1)
-        assert Struct("a"/Byte, Alias("b","a")).build(dict(a=1)) == b"\x01"
-        assert Struct("a"/Byte, Alias("b","a")).sizeof() == 1
-
     def test_bitsinteger(self):
         assert BitsInteger(8).parse(b"\x01\x01\x01\x01\x01\x01\x01\x01") == 255
         assert BitsInteger(8).build(255) == b"\x01\x01\x01\x01\x01\x01\x01\x01"
@@ -537,7 +547,7 @@ class TestCore(unittest.TestCase):
 
     def test_embeddedbitstruct(self):
         d = Struct(
-            "len" / Byte, 
+            "len" / Byte,
             EmbeddedBitStruct("data" / BitsInteger(8)),
         )
         assert d.parse(b"\x08\xff") == Container(len=8)(data=255)
@@ -546,7 +556,7 @@ class TestCore(unittest.TestCase):
 
     def test_bitstruct_from_issue_39(self):
         d = Struct(
-            "len" / Byte, 
+            "len" / Byte,
             EmbeddedBitStruct("data" / BitsInteger(lambda ctx: ctx._.len)),
         )
         assert d.parse(b"\x08\xff") == Container(len=8)(data=255)
@@ -625,7 +635,7 @@ class TestCore(unittest.TestCase):
         assert raises(Select("a"/Int32ub, "b"/Int16ub, "c"/Int8ub, includename=True).build, (("d", 7))) == SelectError
         assert raises(Select(Byte).sizeof) == SizeofError
 
-    @pytest.mark.xfail(not supportskwordered, reason="ordered kw was introduced in 3.6 and pypy")
+    @pytest.mark.xfail(not supportskwordered, reason="ordered kw was introduced in 3.6")
     def test_select_kwctor(self):
         st = Select(a=Int8ub, b=Int16ub, c=Int32ub)
         assert st.parse(b"\x01\x02\x03\x04") == 0x01
@@ -661,15 +671,17 @@ class TestCore(unittest.TestCase):
         assert Union("a"/Bytes(2), "b"/Int16ub, buildfrom="a").build(dict(a=b"zz",b=5))  == b"zz"
         assert Union("a"/Bytes(2), "b"/Int16ub, buildfrom="b").build(dict(a=b"zz",b=5))  == b"\x00\x05"
         assert Union("a"/Bytes(2), "b"/Int16ub, buildfrom=Pass).build({}) == b""
-        assert raises(Union(Pass, buildfrom=123).parse, b"") == IndexError
-        assert raises(Union(Pass, buildfrom="missing").build, {}) == IndexError
+        assert raises(Union(Pass, buildfrom=123).parse, b"") == KeyError
+        assert raises(Union(Pass, buildfrom="missing").parse, b"") == KeyError
+        assert raises(Union(Pass, buildfrom=123).build, {}) == IndexError
+        assert raises(Union(Pass, buildfrom="missing").build, {}) == KeyError
 
         assert raises(Union(Byte).sizeof) == SizeofError
         assert raises(Union(VarInt).sizeof) == SizeofError
         assert Union(Byte, VarInt, buildfrom=0).sizeof() == 1
         assert raises(Union(Byte, VarInt, buildfrom=1).sizeof) == SizeofError
         assert raises(Union(Pass, buildfrom=123).sizeof) == IndexError
-        assert raises(Union(Pass, buildfrom="missing").sizeof) == IndexError
+        assert raises(Union(Pass, buildfrom="missing").sizeof) == KeyError
         assert raises(Union(Pass, buildfrom=this.missing).sizeof) == SizeofError
 
         assert (Union("b"/Int16ub) >> Byte).parse(b"\x01\x02\x03") == [Container(b=0x0102),0x01]
@@ -685,20 +697,28 @@ class TestCore(unittest.TestCase):
         assert raises(Union("a"/Int16ub, Embedded(Struct("b"/Int8ub, "c"/Int8ub)), buildfrom=1).build, dict(b=0x01)) == KeyError
         assert raises(Union("a"/Int16ub, Embedded(Struct("b"/Int8ub, "c"/Int8ub)), buildfrom=1).build, dict()) == KeyError
 
-    @pytest.mark.xfail(not supportskwordered, reason="ordered kw was introduced in 3.6 and pypy")
+    @pytest.mark.xfail(not supportskwordered, reason="ordered kw was introduced in 3.6")
     def test_union_kwctor(self):
         st = Union(a=Int8ub, b=Int16ub, c=Int32ub)
         assert st.parse(b"\x01\x02\x03\x04") == Container(a=0x01,b=0x0102,c=0x01020304)
         assert st.build(Container(c=0x01020304)) == b"\x01\x02\x03\x04"
 
+    def test_union_issue_348(self):
+        u = Union(
+            Int8=Prefixed(Int16ub, Int8ub[:]),
+            Int16=Prefixed(Int16ub, Int16ub[:]),
+            Int32=Prefixed(Int16ub, Int32ub[:]),
+            buildfrom=0)
+        assert u.parse(b'\x00\x04\x11\x22\x33\x44') == {'Int16': [4386, 13124], 'Int32': [287454020], 'Int8': [17, 34, 51, 68]}
+
     def test_prefixed(self):
-        assert Prefixed(Byte, Int16ul).parse(b"\x02\xff\xffgarbage") == 65535
+        assert Prefixed(Byte, Int16ul).parse(b"\x02\xff\xff??????") == 65535
         assert Prefixed(Byte, Int16ul).build(65535) == b"\x02\xff\xff"
         assert Prefixed(Byte, Int16ul).sizeof() == 3
-        assert Prefixed(VarInt, GreedyBytes).parse(b"\x03abcgarbage") == b"abc"
+        assert Prefixed(VarInt, GreedyBytes).parse(b"\x03abc??????") == b"abc"
         assert Prefixed(VarInt, GreedyBytes).build(b"abc") == b'\x03abc'
         assert Prefixed(Byte, Int64ub).sizeof() == 9
-        assert Prefixed(Byte, Sequence(Peek(Byte), Int16ub, GreedyBytes)).parse(b"\x02\x00\xffgarbage") == [0,255,b'']
+        assert Prefixed(Byte, Sequence(Peek(Byte), Int16ub, GreedyBytes)).parse(b"\x02\x00\xff????????") == [0,255,b'']
         assert raises(Prefixed(VarInt, GreedyBytes).sizeof) == SizeofError
 
     def test_compressed_zlib(self):
@@ -712,7 +732,7 @@ class TestCore(unittest.TestCase):
         assert len(d.build(zeros)) < 50
         assert raises(d.sizeof) == SizeofError
 
-    @pytest.mark.xfail(PY < (3,2), reason="gzip was added in 3.2")
+    @pytest.mark.xfail(PY < (3,2), raises=AttributeError, reason="gzip was added in 3.2")
     def test_compressed_gzip(self):
         zeros = bytes(10000)
         d = Compressed(GreedyBytes, "gzip")
@@ -735,7 +755,7 @@ class TestCore(unittest.TestCase):
         assert len(d.build(zeros)) < 50
         assert raises(d.sizeof) == SizeofError
 
-    @pytest.mark.xfail(PY < (3,3), reason="lzma module was added in 3.3")
+    @pytest.mark.xfail(PY < (3,3), raises=ImportError, reason="lzma module was added in 3.3")
     def test_compressed_lzma(self):
         zeros = bytes(10000)
         d = Compressed(GreedyBytes, "lzma")
@@ -826,10 +846,6 @@ class TestCore(unittest.TestCase):
         )
         common(st, b"\x05\x09\x00", Container(value=5)(next=Container(value=9)(next=Container(value=0)(next=None))), SizeofError)
 
-    def test_unknown(self):
-        assert Struct("length" / Byte, "inner" / Struct("inner_length" / Byte, "data" / Bytes(lambda ctx: ctx._.length + ctx.inner_length))).parse(b"\x03\x02helloXXX") == Container(length=3)(inner=Container(inner_length=2)(data=b"hello"))
-        assert Struct("length" / Byte, "inner" / Struct("inner_length" / Byte, "data" / Bytes(lambda ctx: ctx._.length + ctx.inner_length))).sizeof(Container(inner_length=2)(_=Container(length=3))) == 7
-
     def test_noneof(self):
         assert NoneOf(Byte,[4,5,6,7]).parse(b"\x08") == 8
         assert raises(NoneOf(Byte,[4,5,6,7]).parse, b"\x06") == ValidationError
@@ -848,10 +864,28 @@ class TestCore(unittest.TestCase):
         assert Check(this.x == 255).parse(b"", Container(x=255)) == None
         assert Check(this.x == 255).build(None, Container(x=255)) == b""
         assert Check(this.x == 255).sizeof() == 0
-
         assert Check(len_(this.a) == 3).parse(b"", Container(a=[1,2,3])) == None
         assert Check(len_(this.a) == 3).build(None, Container(a=[1,2,3])) == b""
 
+    def test_stopif(self):
+        st = Struct('x'/Byte, StopIf(this.x == 0), 'y'/Byte)
+        common(st, b"\x00", Container(x=0))
+        common(st, b"\x01\x02", Container(x=1,y=2))
+
+        seq = Sequence('x'/Byte, StopIf(this.x == 0), 'y'/Byte)
+        common(seq, b"\x00", [0])
+        common(seq, b"\x01\x02", [1,None,2])
+
+        r = GreedyRange(FocusedSeq(0, 'x'/Byte, StopIf(this.x == 0)))
+        assert r.parse(b"\x00???") == []
+        assert r.parse(b"\x01\x00???") == [1]
+        assert r.build([]) == b""
+        assert r.build([0]) == b"\x00"
+        assert r.build([1]) == b"\x01"
+        assert r.build([1,0,2]) == b"\x01\x00"
+        # common(r, b"\x00???", [], eq=0)
+        # common(r, b"\x01\x00???", [1], eq=0)
+
     def test_hex(self):
         assert Hex(GreedyBytes).parse(b"abcd") == b"61626364"
         assert Hex(GreedyBytes).build(b"61626364") == b"abcd"
@@ -982,6 +1016,19 @@ class TestCore(unittest.TestCase):
         assert Enum(Byte, q=3,r=4,t=5, default=Pass).build(9) == b"\x09"
         assert Enum(Byte, q=3,r=4,t=5).sizeof() == 1
 
+    def test_from_issue_298(self):
+        st = Struct(
+            "ctrl"/Enum(Bytes(1),
+                default = b'\x02',
+                STX = b'\x02',
+                ACK = b'\x06',
+                NAK = b'\x15',
+            ),
+            "optional"/If(this.ctrl == "NAK", Byte),
+        )
+        assert st.parse(b"\x15\xff") == Container(ctrl='NAK')(optional=255)
+        assert st.parse(b"\x02") == Container(ctrl='STX')(optional=None)
+
     def test_flagsenum(self):
         assert FlagsEnum(Byte, a=1,b=2,c=4,d=8,e=16,f=32,g=64,h=128).parse(b'\x81') == FlagsContainer(a=True,b=False,c=False,d=False,e=False,f=False,g=False,h=True)
         assert FlagsEnum(Byte, a=1,b=2,c=4,d=8,e=16,f=32,g=64,h=128).build(FlagsContainer(a=True,b=False,c=False,d=False,e=False,f=False,g=False,h=True)) == b'\x81'
@@ -1089,7 +1136,7 @@ class TestCore(unittest.TestCase):
         assert IpAddress.build("127.1.2.3") == b"\x7f\x01\x02\x03"
         assert IpAddress.sizeof() == 4
 
-        IpAddress = ExprAdapter(Byte[4], 
+        IpAddress = ExprAdapter(Byte[4],
             encoder = lambda obj,ctx: list(map(int, obj.split("."))),
             decoder = lambda obj,ctx: "{0}.{1}.{2}.{3}".format(*obj), )
 
@@ -1278,7 +1325,7 @@ class TestCore(unittest.TestCase):
         NumVertices = Bitwise(Aligned(8, Struct(
             'numVx4' / BitsInteger(4),
             'numVx8' / If(this.numVx4 == 0, BitsInteger(8)),
-            'numVx16' / If(this.numVx4 == 0 and this.numVx8 == 255, BitsInteger(16)),
+            'numVx16' / If(this.numVx4 == 0 & this.numVx8 == 255, BitsInteger(16)),
         )))
         common(NumVertices, b'\x02\x30', Container(numVx4=0, numVx8=35, numVx16=None))
 
@@ -1293,14 +1340,12 @@ class TestCore(unittest.TestCase):
         common(testBit, b'ab', Container(a=97, b=98))
         common(testByte, b'ab', Container(a=97, b=98))
 
-    @pytest.mark.xfail(raises=AssertionError, reason="makes no sense")
-    def test_from_issue_246_third(self):
         NumVertices = Union(
-            'numVx4' /  Bitwise(Aligned(8, Struct('num'/ BitsInteger(4)))),
-            'numVx8' /  Bitwise(Aligned(8, Struct('num'/ BitsInteger(12)))),
-            'numVx16' / Bitwise(Aligned(8, Struct('num'/ BitsInteger(28)))),
+            'numVx4' / Bitwise(Aligned(8, Struct('num'/ BitsInteger(4) ))),
+            'numVx8' / Bitwise(Aligned(8, Struct('num'/ BitsInteger(12)))),
+            'numVx16'/ Bitwise(Aligned(8, Struct('num'/ BitsInteger(28)))),
         )
-        common(NumVertices, b'\x01\x34\x56\x70', Container(numVx4=Container(num=0))(numVx8=Container(num=19))(numVx16=Container(num=1262951)))
+        assert NumVertices.parse(b'\x01\x34\x56\x70') == Container(numVx4=Container(num=0))(numVx8=Container(num=19))(numVx16=Container(num=1262951))
 
     def test_from_issue_244(self):
         class AddIndexes(Adapter):
@@ -1313,4 +1358,71 @@ class TestCore(unittest.TestCase):
 
         assert AddIndexes(Struct("num"/Byte)[4]).parse(b"abcd") == [Container(num=97)(index=0),Container(num=98)(index=1),Container(num=99)(index=2),Container(num=100)(index=3),]
 
+    def test_from_issue_269(self):
+        st = Struct("enabled" / Byte, If(this.enabled, Padding(2)))
+        assert st.build(dict(enabled=1)) == b"\x01\x00\x00"
+        assert st.build(dict(enabled=0)) == b"\x00"
+        st = Struct("enabled" / Byte, "pad" / If(this.enabled, Padding(2)))
+        assert st.build(dict(enabled=1)) == b"\x01\x00\x00"
+        assert st.build(dict(enabled=0)) == b"\x00"
 
+    def test_hanging_issue_280(self):
+        st = BitStruct('a'/BitsInteger(20), 'b'/BitsInteger(12))
+        assert raises(st.parse, b'\x00') == IOError
+
+    def test_nonbytes_checksum_issue_323(self):
+        st = Struct(
+            "vals" / Byte[2],
+            "checksum" / Checksum(
+                Byte, lambda vals: sum(vals) & 0xFF, this.vals)
+        )
+        assert st.parse(b"\x00\x00\x00") == Container(vals=[0, 0])(checksum=0)
+        assert raises(st.parse, b"\x00\x00\x01") == ChecksumError
+
+    def test_from_issue_324(self):
+        d = Struct(
+            "vals" / Prefixed(Byte, RawCopy(
+                Struct("a" / Byte[2]),
+            )),
+            "checksum" / Checksum(
+                Byte,
+                lambda data: sum(iterateints(data)) & 0xFF,
+                this.vals.data
+            ),
+        )
+        assert d.build(dict(vals=dict(value=dict(a=[0,1])))) == b"\x02\x00\x01\x01"
+        assert d.build(dict(vals=dict(data=b"\x00\x01"))) == b"\x02\x00\x01\x01"
+
+    def test_embeddedif_issue_296(self):
+        st = 'BuggedStruct' / Struct(
+            'ctrl' / Bytes(1),
+            Probe(),
+            Embedded(If(
+                this.ctrl == b'\x02',
+                Struct('etx' / Const(b'\x03')),
+            )),
+            Probe(),
+        )
+        p1 = st.parse(b'\x02\x03')
+        p3 = st.parse(b'\x06')
+
+    def test_embeddedswitch_issue_312(self):
+        st = Struct(
+            'name'/CString(encoding="utf8"),
+            Embedded(If(len_(this.name) > 0,
+                Struct('index'/Byte),
+            )),
+        )
+        assert st.parse(b'bob\x00\x05') == Container(name='bob')(index=5)
+        assert st.parse(b'\x00') == Container(name='')
+
+    @pytest.mark.xfail(strict=True, reason="this cannot work, Struct checks flagembedded before building")
+    def test_embeddedswitch_issue_312_cannotwork(self):
+        st = Struct(
+            'name'/CString(encoding="utf8"),
+            If(len_(this.name) > 0,
+                Embedded(Struct('index'/Byte)),
+            ),
+        )
+        assert st.parse(b'bob\x00\x05') == Container(name='bob')(index=5)
+        assert st.parse(b'\x00') == Container(name='')
